#include <assert.h>
#include <QMenu>
#include "PluginManager.h"
#include "PluginServices.h"
#include "PluginList.h"
#include "PluginSettingDialog.h"
#include "TreeItem.h"
#include "MainWidget.h"
#include "Globals.h"
#define NO_PLUGIN -1

// workaround:leak in plugin may cause segfault https://bugreports.qt-project.org/browse/QTBUG-39070
#define NO_PLUGIN_UNLOAD

PluginManager* PluginManager::single                  = 0;
bool           PluginManager::          disableAlways = false;

PluginManager*
PluginManager::getInstance()
{
    if ( single == 0 )
    {
        single = new PluginManager();
    }
    return single;
}

PluginManager::PluginManager()
{
    cube         = 0;
    tabManager   = 0;
    pluginMenu   = 0;
    cubeFileName = "";

    currentContextFreePlugin = 0;

    if ( disableAlways )
    {
        return;
    }

    enableAllPlugins  = true;
    disableAllPlugins = false;
    getPluginStatus( enableAllPlugins, disableAllPlugins, pluginStatus );

    // Create initial plugin menu structure
    int pluginCount = PluginList::getCubePluginCount();
    if ( pluginCount > 0 )
    {
        pluginMenu = new QMenu( tr( "&Plugins" ) );
        connect( pluginMenu, SIGNAL( aboutToShow() ), this, SLOT( checkPluginMenuStatus() ) );

        QAction* item = new QAction( "Initial activation settings", pluginMenu );
        item->setWhatsThis( "Allows to activate or deactivate plugins independent of the cube file" );
        pluginMenu->addAction( item );
        connect( item, SIGNAL( triggered() ), this, SLOT( showSettingsWidget() ) );

        QMenu* pluginStatusMenu = pluginMenu->addMenu( "Activate/deactivate Plugins" );
        pluginStatusMenu->setWhatsThis( "Allows to activate or deactivate plugins which are available for the current cube file" );
        pluginMenu->addSeparator();

        for ( int index = 0; index < pluginCount; ++index )
        {
            CubePlugin* plugin  = PluginList::getCubePlugin( index );
            QMenu*      subMenu = new QMenu( plugin->name(), pluginMenu );
            subMenu->menuAction()->setVisible( false );
            pluginMenu->addMenu( subMenu );
            pluginMenuList.append( subMenu );
            pluginServicesList.append( 0 );

            // add "active/deactivate plugin" menu item
            QAction* pa = new QAction( plugin->versionLabel(), this );
            pluginActivationAction.append( pa );
            pa->setCheckable( true );
            pa->setChecked( true );
            pa->setStatusTip( tr( "Activate/deactivate plugin" ) );
            pa->setWhatsThis( "If this menu item is checked, the plugin is activated" );
            pa->setData( index );
            connect( pa, SIGNAL( triggered( bool ) ), this, SLOT( enablePlugin( bool ) ) );
            pluginStatusMenu->addAction( pa );
            pa->setVisible( false );
        }
    }
}

void
PluginManager::setMainWindow( QMainWindow* main )
{
    mainWindow = main;
}

void
PluginManager::disablePlugins()
{
    PluginList::disablePlugins();
    disableAlways = true;
}

void
PluginManager::opened( cube::Cube* cube, const QString& file, TabManager* tabManager )
{
    this->tabManager   = tabManager;
    this->cube         = cube;
    this->cubeFileName = file;
}

void
PluginManager::loadPlugins()
{
    if ( cube == 0 || tabManager == 0 )
    {
        return; // not opened
    }
    if ( !disableAlways )
    {
        loadAllPlugins();
    }
}

void
PluginManager::openPlugin( int index )
{
    bool        ok     = true; // plugin has been loaded and is valid
    CubePlugin* plugin = PluginList::getCubePlugin( index );

    /*
       If plugin isn't loaded, load and initialize if. Special case 1st run: All
       available plugins are loaded but not yet initialized.
     */
#ifndef NO_PLUGIN_UNLOAD
    if ( !plugin )
    {
        plugin = PluginList::loadPluginAt( index );
    }
#endif

    if ( plugin && !plugin->isInitialized )
    {
        PluginServices* services = new PluginServices( this, index );
        pluginServicesList[ index ] = services;

        ok                    = plugin->cubeOpened( services );
        plugin->isInitialized = true;
    }

    if ( ok )
    {
        pluginMenuList.at( index )->menuAction()->setVisible( true );
        pluginActivationAction.at( index )->setVisible( true );
    }
    else
    {
        pluginMenuList.at( index )->menuAction()->setVisible( false );
        pluginActivationAction.at( index )->setVisible( false );
        this->closePlugin( index );
    }
}

void
PluginManager::closePlugin( int index )
{
    CubePlugin* plugin = PluginList::getCubePlugin( index );

    if ( plugin == 0 )
    {
        return; // already unloaded
    }

    QMenu* subMenu = pluginMenuList.at( index );
    subMenu->clear(); // clear all plugin related menu items

    PluginServices* service = pluginServicesList[ index ];

    // remove plugin toolbar(s)
    foreach( QToolBar * bar, toolBarHash.values( service ) )
    {
        mainWindow->removeToolBar( bar );
        toolBarHash.remove( service, bar );
    }

    // remove plugin colormaps
    foreach( ColorMap * map, colormaps.values( service ) )
    {
        if ( map == Globals::getColorMap() )
        {
            Globals::setDefaultColorMap();
        }
        colormaps.remove( service, map );
    }

    if ( service )
    {
        service->closeTabs();
        plugin->cubeClosed();
        plugin->isInitialized = false;

        // delete corresponding service
        delete service;
        pluginServicesList[ index ] = 0;
    }

#ifndef NO_PLUGIN_UNLOAD
    PluginList::unloadPluginAt( index );
#endif
}

QMenu*
PluginManager::getPluginMenu()
{
    return pluginMenu;
}

void
PluginManager::loadAllPlugins()
{
    for ( int i = 0; i < PluginList::getCubePluginCount(); ++i )
    {
        QString name = PluginList::getCubePluginName( i );

        bool load = true;
        if ( enableAllPlugins )
        {
            load = true;
        }
        else if ( disableAllPlugins )
        {
            load = false;
        }
        else
        {
            load = pluginStatus.value( name, true ); // load plugin, if not not set to false in settings
        }

        if ( load )
        {
            openPlugin( i );
        }
        else
        {
            closePlugin( i );
        }
        pluginActivationAction[ i ]->setChecked( load );
    }
}

// called if cube file has been closed
void
PluginManager::closed()
{
    for ( int i = 0; i < pluginMenuList.size(); ++i )
    {
        closePlugin( i );
        pluginActivationAction.at( i )->setVisible( false );
    }

    this->tabManager   = 0;
    this->cube         = 0;
    this->cubeFileName = "";

    // reinitialize plugin status to startup settings
    getPluginStatus( enableAllPlugins, disableAllPlugins, pluginStatus );

    // remove internal services
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service && service->index < 0 )
        {
            pluginServicesList.removeOne( service );
        }
    }
    // end remove internal services
}

void
PluginManager::enablePlugin( bool enable )
{
    QAction* act = dynamic_cast<QAction*>( sender() );
    assert( act );

    int index = act->data().toInt();
    if ( enable )
    {
        openPlugin( index  );
    }
    else
    {
        closePlugin( index  );
    }
}

void
PluginManager::showSettingsWidget()
{
    PluginSettingDialog::showSettingsDialog( tabManager );
}

/**
 * checkPluginMenuStatus hides empty plugin menus
 */
void
PluginManager::checkPluginMenuStatus()
{
    foreach( QMenu * menu, pluginMenuList )
    {
        if ( menu->isEmpty() )
        {
            menu->menuAction()->setVisible( false );
        }
        else
        {
            menu->menuAction()->setVisible( true );
        }
    }
}

// ---------- emit signals for plugins -----------------

void
PluginManager::contextMenuIsShown( TreeType type, TreeItem* treeItem )
{
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service )
        {
            emit service->contextMenuIsShown( type, treeItem );
        }
    }
}

void
PluginManager::treeItemIsSelected( TreeType type, TreeItem* treeItem, bool selected )
{
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service )
        {
            if ( selected )
            {
                emit service->treeItemIsSelected( type, treeItem );
            }
            else
            {
                emit service->treeItemIsDeselected( type, treeItem );
            }
        }
    }
}

void
PluginManager::setTabOrder( const QList<DisplayType>& order )
{
    this->order = order;
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service )
        {
            emit service->orderHasChanged( order );
        }
    }
}

void
PluginManager::setGlobalValue( const QString& name, const QVariant& value, PluginServices* sender )
{
    globalValues.insert( name, value );

    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service && service != sender )
        {
            emit service->globalValueChanged( name );
        }
    }
}

QMainWindow*
PluginManager::getMainWindow()
{
    return mainWindow;
}


void
PluginManager::genericUserAction( UserAction action )
{
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service )
        {
            emit service->genericUserAction( action );
        }
    }
}

// ---------- end emit signals for plugins -----------------



/**
 * reads the initial plugin status from settings
 */
void
PluginManager::getPluginStatus( bool& all, bool& none, QHash<QString, bool>& plugins )
{
    all     = enableAllPlugins;
    none    = disableAllPlugins;
    plugins = pluginStatus;
}


/**
 * @brief PluginManager::setPluginStatus
 * Activates or deactivates the available plugins and saves the status
 * @param all all plugins will be activated, if true
 * @param none all plugins will be deactivated, if true
 * @param plugins Hash with pluginName->activated value pairs
 */
void
PluginManager::setPluginStatus( bool& all, bool& none, QHash<QString, bool>& plugins )
{
    enableAllPlugins  = all;
    disableAllPlugins = none;
    pluginStatus      = plugins;

    if ( cube )
    {
        loadAllPlugins();
    }
}

// --- settings for initial activation/deactivation of the plugins

/** this method is only called once if program ends */
void
PluginManager::saveGlobalStartupSettings( QSettings& settings )
{
    if ( disableAlways )
    {
        return;
    }

    QStringList enabledList, disabledList;
    foreach( QString pluginName, pluginStatus.keys() )
    {
        if ( pluginStatus.value( pluginName ) )
        {
            enabledList.append( pluginName );
        }
        else
        {
            disabledList.append( pluginName );
        }
    }

    settings.setValue( "pluginEnabledList", enabledList );
    settings.setValue( "pluginDisabledList", disabledList );
}

/** this method is only called once at program start */
void
PluginManager::loadGlobalStartupSettings( QSettings& settings )
{
    if ( disableAlways )
    {
        return;
    }

    QStringList enabledList, disabledList;
    enabledList  = settings.value( "pluginEnabledList" ).toStringList();
    disabledList = settings.value( "pluginDisabledList" ).toStringList();

    foreach( QString plugin, enabledList )
    {
        pluginStatus.insert( plugin, true );
    }
    foreach( QString plugin, disabledList )
    {
        pluginStatus.insert( plugin, false );
    }

    enableAllPlugins  = false; // always false, since the activation status of each plugin is saved
    disableAllPlugins = false;
}

// ----- implementation of SettingsHandler interface to save settings of each plugin  -----------

QString
PluginManager::settingName()
{
    return "plugins";
}

void
PluginManager::saveGlobalSettings( QSettings& settings )
{
    // save settings for each registered pluginService
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service && service->settingsHandler )
        {
            settings.beginGroup( service->settingsHandler->settingNameCompact() );
            service->settingsHandler->saveGlobalSettings( settings );
            settings.endGroup();
        }
    }
}

void
PluginManager::loadGlobalSettings( QSettings& settings )
{
    // load settings for pluginService with registered settingsHandler
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service && service->settingsHandler )
        {
            settings.beginGroup( service->settingsHandler->settingNameCompact() );
            service->settingsHandler->loadGlobalSettings( settings );
            settings.endGroup();
        }
    }
}

void
PluginManager::saveExperimentSettings( QSettings& settings )
{
    // save settings for each registered pluginService
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service && service->settingsHandler )
        {
            settings.beginGroup( service->settingsHandler->settingNameCompact() );
            service->settingsHandler->saveExperimentSettings( settings );
            settings.endGroup();
        }
    }
}

void
PluginManager::loadExperimentSettings( QSettings& settings )
{
    // load settings for each registered pluginService
    foreach( PluginServices * service, pluginServicesList )
    {
        if ( service && service->settingsHandler )
        {
            settings.beginGroup( service->settingsHandler->settingNameCompact() );
            service->settingsHandler->loadExperimentSettings( settings );
            settings.endGroup();
        }
    }
}



PluginServices*
PluginManager::createInternalService()
{
    PluginServices* service = new PluginServices( this, NO_PLUGIN );
    pluginServicesList.append( service );
    return service;
}

// ============== context free plugins ====================================

#include "ContextFreeServices.h"

/**
 * @brief PluginManager::setContextFreePluginMenu
 * @param menu the menu which is filled with the labels of context free plugins
 */
void
PluginManager::setContextFreePluginMenu( QMenu* menu )
{
    int pluginCount = PluginList::getContextFreePluginCount();
    if ( pluginCount == 0 )
    {
        menu->setEnabled( false );
    }
    else
    {
        for ( int index = 0; index < pluginCount; ++index )
        {
            ContextFreePlugin* plugin = PluginList::getContextFreePlugin( index );
            QAction*           pa     = new QAction( plugin->versionLabel(), this );
            menu->addAction( pa );
            pa->setData( index );
            connect( pa, SIGNAL( triggered() ), this, SLOT( startPlugin() ) );
        }
        menu->addSeparator();
        closeContextAction = new QAction( "Close current Plugin", this );
        closeContextAction->setEnabled( false );
        menu->addAction( closeContextAction );
        connect( closeContextAction, SIGNAL( triggered() ), this, SLOT( closeContextFreePlugin() ) );
    }
}

/** slot for menu action */
void
PluginManager::startPlugin()
{
    // close active plugin, if exists
    if ( currentContextFreePlugin )
    {
        currentContextFreePlugin->closed();
    }

    // open chosen plugin
    QAction*           action = qobject_cast<QAction*>( sender() );
    int                index  = action->data().toInt();
    ContextFreePlugin* plugin = PluginList::getContextFreePlugin( index ); // selected plugin from menu
    currentContextFreePlugin = plugin;
    plugin->opened( ContextFreeServices::getInstance() );
    closeContextAction->setEnabled( true );
}

/** closes active context free plugin */
void
PluginManager::closeContextFreePlugin()
{
    if ( currentContextFreePlugin )
    {
        currentContextFreePlugin->closed();

        // remove layout items of the previously started plugin
        QWidget* w      = ContextFreeServices::getInstance()->getWidget();
        QLayout* layout = w->layout();
        while ( layout->count() > 0 )
        {
            QWidget* child = layout->itemAt( 0 )->widget();
            delete child;
        }
        delete layout;

        closeContextAction->setEnabled( false );
    }
}
